#!/usr/bin/perl -w
# Copyright Mark Zealey 2002. Released under the GPL.
use strict;
use MSN;
use Curses;
use Text::Wrap;
use IO::Handle;
use POSIX qw/strftime/;

my $RESIZE_PROG = '/usr/X11R6/bin/resize';	# Program to get new window size
my $CONFIG = "$ENV{HOME}/.jMSN.conf";		# Config file location
my $MSN_VERSION;

# Go thru our command-line args
while(defined($_ = shift)) {
	$_ eq '-f' and $CONFIG = shift;
	$_ eq '-v' and $MSN_VERSION = shift;
}

if(!$MSN_VERSION) {
	open V, 'VERSION' or die "Can't open VERSION, $!";
	chomp($MSN_VERSION = <V>);
	close V
}

my %conf;
&read_config;
$conf{time} ||= '';
$conf{ping} ||= 60;
$conf{ping_count} ||= 4;
$conf{idle} = 600 if !exists $conf{idle};
my $MAX_FUSER_LEN = $conf{fuser_len} || 0;
my $MAX_BUFFER_LEN = $conf{buffer_len} || 1000;
my $MAX_LINE_LEN = $conf{input_buf} || 50;

my $time_hires;
eval 'use Time::HiRes qw/time/';
$time_hires = 1 unless $@;

my $time = time;			# The current time
my $startup_time = $time;		# Time we started
my $session = -1;			# Number of the current session
my %sessions;				# Info about all sessions
my ($prompt, $ask);
my $last_msg = '';
my $escape;				# Status of the escape key
my $disconnect = 0;			# Disconnect when this is 1
my $hide_input = 0;			# Hide input from being displayed?
my $last_type = $time;			# Contains the last time the user typed
my $last_ping = '??';			# Time taken for the last ping
my $ping_last;				# When we sent our last ping
my $ping_got = 0;			# When we got our last ping
my @lines;				# List of the lines we've read from the
					# input
my $line_ptr = 0;			# Pointer to the elem in @lines we are
					# ref'ing
my $char_ptr = -1;			# Current pos we are on the line
my $replace_on = 0;			# Do we replace or insert stuff?
my $needs_resize = 0;			# Need to resize the windows etc?
my $idle = 0;				# We idle?
my $last_user = 0;			# The last time we updated the user box
my $last_active = 0;			# The time of the last active user
my $ping_in_progress = 0;		# We waiting for a ping response?
my @users;				# List of users in current window
my $away_msg = undef;			# Away message
my (%active, %vactive);			# Active and very active windows
my $user_up = 1;			# Update user window?
my $stat_up = 1;			# Update status window
my $in_up = 1;				# Update prompt window
my $us_len = 1;				# Rows in the user window

my ($us_line, $in_line, $stat_line, $MAGENTA, $RED, $BLUE);

# What to put for the types of arguments that are mentioned below.
my %text_translate = (
		WINDOW => 'window',
		SESSION => 'session',
		COMMAND => 'command',
		FNUM => 'file number',
		FILE => 'filename',
		USER => 'username',
		TEXT => 'text',
		STATE => 'status',
	);

# This is the big array, defining all the /command's we can handle. It's fairly
# simple, in each array element, the following info is contained:
#
# element | description
# --------+---------------------------------------------------------------------
#       1 | An array-ref to an array of names this function is called. The first
#         | element in this sub-array is the name of the function (without the
#         | handle_ bit on the front) that will be called when any of the names
#         | in the list is typed
# --------+---------------------------------------------------------------------
#       2 | An array-ref of options. Each option must be preceeded by its
#         | required-ness, ie, a string containing some of:
#         |             REQ (required)
#         |             OPT (optional)
#         |             MUL (multiple)
# --------+---------------------------------------------------------------------
#       3 | A string which says what type of arg this is. This can be one of:
#         |             WINDOW - window number
#         |             SESSION - session number
#         |             COMMAND - a command
#         |             FNUM - file transfer number
#         |             FILE - filename
#         |             USER - username
#         |             TEXT - any text (not completed)
#         |             STATE - A state
my @funcs = (
		[
			[qw/away/],
			[
				'OPT MUL' => 'TEXT'
			],
			'Change your status to "away" and sets the away ' .
			'message to [1]'
		], [
			[qw/hide hidden/],
			[],
			'Change your status to "hidden"'
		], [
			[qw/status/],
			[
				'REQ' => 'STATE',
				'OPT MUL' => 'TEXT'
			],
			'Change your status to "[1]" and away message to [2]'
		], [
			[qw/online/],
			[],
			'Change your status to "online"'
		], [
			[qw/window session/],
			[
				'REQ' => 'WINDOW'
			],
			'Switches to session [1]'
		], [
			[qw/msg/],
			[
				'REQ' => 'SESSION',
				'REQ MUL' => 'TEXT'
			],
			'Send [2] to session [1]'
		], [
			[qw/crash/],
			[
				'OPT' => 'SESSION'
			],
			'Crashes all the windows msn clients in the specified ' .
			'session'
		], [
			[qw/quit exit/],
			[],
			'Quits MSN'
		], [
			[qw/part bye/],
			[
				'OPT' => 'SESSION'
			],
			'Leave [1], or the current one if none is specified'
		], [
			[qw/add/],
			[
				# Text because we do not have the user already
				'REQ MUL' => 'TEXT'
			],
			'Attempts to add [1] to your buddies list, so you ' .
			'will be informed when they are online etc...'
		], [
			[qw/del/],
			[
				'REQ MUL' => 'USER'
			],
			'Attempts to remove [1] from your buddies list'
		], [
			[qw/authorize/],
			[
				'REQ MUL' => 'USER'
			],
			'Authorizes [1] to add you to their buddie list'
		], [
			[qw/deauthorize/],
			[
				'REQ MUL' => 'USER'
			],
			'Un-authorizes [1] to add you to their buddie list'
		], [
			[qw/block ban/],
			[
				'REQ MUL' => 'USER'
			],
			'Attempts to block [1], ie stops them from talking to '.
			'you'
		], [
			[qw/unblock unban/],
			[
				'REQ MUL' => 'USER'
			],
			'Attempts to unblock [1], ie lets them talk to you ' .
			'again'
		], [
			[qw/chat/],
			[
				'REQ MUL' => 'USER'
			],
			'Attempts to start a chat with the [1]'
		], [
			[qw/clear/],
			[
				'OPT' => 'WINDOW'
			],
			'Clears the screen of window [1] (or the current one '.
			'if none specified'
		], [
			[qw/invite/],
			[
				'REQ' => 'USER',
				'OPT' => 'SESSION'
			],
			'Invites [1] to the specified (or current) chat session'
		], [
			[qw/change_fuser setname setfuser nick/],
			[
				'REQ MUL' => 'TEXT'
			],
			'Changes your "friendly" username to [1]'
		], [
			[qw/file_get get/],
			[
				'REQ' => 'FNUM'
			],
			'Attempts to get the file that is waiting on [1]'
		], [
			[qw/file_send send/],
			[
				'REQ' => 'SESSION',
				'REQ' => 'FILE'
			],
			'Attempts to send [2] to the (single) user on chat ' .
			'session [1]'
		], [
			[qw/file_noget noget cancel/],
			[
				'REQ' => 'FNUM'
			],
			'Attempts to cancel [1]'
		], [
			[qw/help/],
			[
				'OPT' => 'COMMAND'
			],
			'Displays this help, or help on [1]'
		], [
			[qw/ping/],
			[],
			'Pings yourself to check the connection is still alive'
		], [
			[qw/uptime/],
			[],
			'Displays how long the client has been running for'
		], [
			[qw/version/],
			[],
			'Gives you info about the client'
		], [
			[qw/refresh redraw/],
			[],
			'Refreshes the screen after some other program fucked '.
			'it up'
		], [
			[qw/list/],
			[
				'OPT MUL' => 'STATE'
			],
			'Gives you a list of people who are not offline at the'.
			' moment, and their states'
		], [
			[qw/destroy/],
			[
				'OPT' => 'WINDOW'
			],
			'Destroys the window number [1]'
		], [
			[qw/eval/],
			[
				'REQ MUL' => 'TEXT'
			],
			'Evaluates the specified perl code and outputs any ' .
			'output'
		], [
			[qw/exec/],
			[
#				'OPT' => '-o',
				'REQ MUL' => 'TEXT'
			],
			'Evaluates the specified sh code and outputs any ' .
			'output to the current session if -o is specified, ' .
			'otherwise echos to the current screen.'
		], [
			[qw/save/],
			[
				'REQ MUL' => 'TEXT'		# May not exist
			],
			'Saves the contents of the current buffer to [1]'
		], [
			[qw/show_fuser/],
			[
				'REQ' => 'USER'
			],
			'Outputs the full fusername of [1]'
		]
	);

my %funcs;

# We now map all the possible functions into the keys hash, which conveniantly
# all point at the 'primary' function, ie, the one we want called.. :)
for(@funcs) {
	my $pri = $_->[0][0];

	for my $t (@{$_->[0]}) {
		$funcs{$t} = [$pri, $_]
	}
}

# Set up our select bitmap
my $stdin = '';
vec($stdin, fileno STDIN, 1) = 1;

&init_screen;
new_win(0);
change_win(0);

# Restore the screen to original state if we get killed
$SIG{$_} = sub { &quit if $disconnect; $disconnect = 1 } for qw/KILL STOP TERM/,
								qw/INT/;

$SIG{WINCH} = sub { $needs_resize = 1 };

# Set a signal handler to stop the screen being messed up when something bad
# happens
$SIG{__WARN__} = sub { out(undef, @_) };
$SIG{__DIE__} = sub { out(undef, @_); &quit };

&handle_version;
out(undef, 'Type /help to display help');
out(undef, 'Attempting to connect...');

&update_prompt;
&screen_update;

# Open the log file
open F, '>>log' or die "Can't open log: $!";
F->autoflush;

# Start the MSN module
my $msn = MSN->new(Debug => \*F) or &quit;

# Register this list of callbacks
my @callbacks = (
			qw/connect message new_mail open_chat disconnected/,
			qw/close_chat user_types join part change_state out/,
			qw/file_invite file_start error unknown_user/,
			qw/auth_success deauth_success add_success del_success/,
			qw/block_success unblock_success add del ping/,
			qw/state_success fuser_success file_send_end/,
			qw/file_get_end folder_update/,
		);

eval '$msn->add_callback($_, sub { &do_' . $_ . '; &screen_update } )' for @callbacks;

$conf{user} ||= ask(0, 'Please enter your MSN username');
$conf{pass} ||= ask(1, 'Please enter your MSN password');

my %opts = (
		User => $conf{user},
		Pass => $conf{pass},
	);

$opts{Host} = $conf{host} if exists $conf{host};
$opts{Port} = $conf{port} if exists $conf{port};
$opts{FUser} = $conf{fuser} if exists $conf{fuser};

# Try to connect!
if(!defined $msn->connect(%opts)) {
	error("Can't connect: $msn->{Error}");
	&quit
}

while(1) {
	$disconnect and &quit;
#	$needs_resize and &resize_term;

	&screen_update;

	my ($poll_func, $interval);

	if(!$conf{ping} and $idle) {
		$interval = undef;		# Don't care
		$poll_func = \&poll_idle
	} elsif($ping_in_progress) {
		$interval = 1;
		$poll_func = \&poll_pip
	} elsif($last_active) {
		$interval = 5;			# 10 / 2
		$poll_func = \&poll_active
	} elsif($conf{idle} and $conf{idle} < $conf{ping}) {
		$interval = $conf{idle};
		$poll_func = \&poll_inact
	} else {				# Just ping
		$interval = $conf{ping};	# Doesn't matter much
		$poll_func = \&poll_ping
	}

	# Got anything to do with MSN?
	my $fh = $msn->do_stuff($poll_func, $interval, $stdin);
	&quit unless defined $fh;

	# Save time in a variable now so we don't waste a lot of system time
	# asking for the time...
	$time = time;

	if($fh->[0] and $fh->[0] eq $stdin) {		# User typed something
		my $t = &in;
		func($t) if defined $t
	} elsif($conf{idle} and $msn->get_state eq 'online'
					and $time - $last_type >= $conf{idle}) {
		$msn->change_my_state('idle');
		out(undef, "Starting auto-idle after $conf{idle} seconds");
		$idle = 1
	}

	# Check to see if any of our users have gone idle from active..
	&update_win if $last_active and $last_active + 5 < $time;

	# Lag-timer
	if($conf{ping} and defined $ping_last) {
		my $diff = $time - $ping_last;

		# If we didn't get a ping after we sent the last, and we're 1
		# second more than when we sent the last, start incrementing
		# the lag counter till we get one back.. (We might have
		# disconnected)
		if($ping_got < $ping_last and $diff >= $conf{ping_count}) {
			$last_ping = $time_hires ? sprintf('%0.3f', $diff)
						 : int $diff;
			&update_status
		}

		&handle_ping if $diff >= $conf{ping}
	}
}

sub quit {
	out(undef, 'Disconnecting');
	&handle_uptime;
	$msn->disconnect if defined $msn;
	close F;
	$in_line->erase;
	&screen_update;
	endwin;				# Close down the windowing system
	exit
}

# Now verifys the command structure too
sub func {
	my $line = shift;
	my ($command, @args);

	# Get the command and it's type
	if($line =~ s!^/(\S+)\s*!!) {
		return unless do_cmd(\$command, $1);
		$command = $funcs{$command}[0]		# Get the principle cmd
	} else {
		$line =~ s!^/\s*!!;		# / /foo
		$line = "$session $line";
		$command = 'msg'
	}

	# Now get and verify the args:
	my (@req, $type);

	for(@{$funcs{$command}[1][1]}) {
		unless(@req) {
			@req = split ' ';
			next
		}

		my $mul = is_in('MUL', @req);

		$type = $_;

		# Get the next argument
		do {
			if($type eq 'TEXT' and $mul) {	# XXX: Special case
				push @args, $line;
				$line = '';
				last
			}

			if($line !~ s/^(\S+)\s*//) {
				if(is_in('REQ', @req) and !is_in('MUL', @req)) {
					error('Required parameter',
						scalar @args, 'is missing')
				}

				last
			}

			my $bit = $1;
			my $err = sub { error("Invalid $text_translate{$type}",
								"'$bit'") };

			# Now do the various type checking (XXX, do we want
			# partial command completion here?)
			if($type eq 'COMMAND') {
				$bit = lc $bit;
				return &$err if !exists $funcs{$bit}
			} elsif($type eq 'WINDOW') {
				return &$err if !exists $sessions{$bit}
			} elsif($type eq 'STATE') {
				$bit = lc $bit;
				return &$err if !exists $MSN::STATES{$bit}
			} elsif($type eq 'SESSION') {
				return &$err if !$msn->sess_exists($bit)
			} elsif($type eq 'USER') {
				return &$err if !$msn->get_user($bit)
			} elsif($type eq 'FILE') {
				return &$err if !-f $bit
			} elsif($type eq 'FNUM') {
				# XXX more checking here.
				return &$err if $bit =~ /\D/
			} elsif($type eq 'TEXT') {
				# No error checking here.
			} else {
				error("Internal: Unknown type '$type'")
			}

			push @args, $bit
		} while($mul);

		@req = ()
	}

	return error("Trailing garbage on command ($line)") unless $line eq '';

	eval '&handle_' . $command . '(@args)';
	error($@) if $@
}

################################################################################
# Polling subroutines
################################################################################

sub poll_ping { time - $time < $conf{ping} }
sub poll_inact { time - $time < $conf{idle} }
sub poll_pip { time - $time < 0.8 }
sub poll_active { time - $last_active < 10 }
sub poll_idle { 1 }

################################################################################
# Handle bits
################################################################################

sub handle_quit		{ &quit }

sub handle_add		{ $msn->add_contact($_)		for @_ }
sub handle_del		{ $msn->del_contact($_)		for @_ }
sub handle_block	{ $msn->block_contact($_)	for @_ }
sub handle_unblock	{ $msn->unblock_contact($_)	for @_ }
sub handle_authorize	{ $msn->authorize_contact($_)	for @_ }
sub handle_deauthorize	{ $msn->deauthorize_contact($_)	for @_ }

sub handle_away		{ handle_status('away', @_) }
sub handle_online	{ handle_status('online') }
sub handle_hide		{ handle_status('hidden') }
sub handle_status {
	$msn->change_my_state(shift);
	$away_msg = "@_" || undef;
	&update_status
}

sub handle_change_fuser	{ $msn->change_fuser(shift) }
sub handle_chat		{ $msn->start_ring(@_) }

sub handle_clear {
	my $win = shift;
	$win = $session if !$win;

	$sessions{$win}{Screen} = [];
	$sessions{$win}{Scroll} = 0;
	make_active($win);

	&update_win if $win == $session;
	&update_status if $win != $session;
}

sub handle_version {
	out(undef, "jMSN version $MSN_VERSION by Mark Zealey",
							'(mark@zealos.org)');
}

sub handle_uptime {
	my ($sec, $min, $hours, $days);

	$days = $time - $startup_time;

	$sec = $days % 60;
	$days = int($days / 60);
	$min = $days % 60;
	$days = int($days / 60);
	$hours = $days % 24;
	$days = int($days / 24);

	out(undef, "Current uptime: ${days}d ${hours}h ${min}m ${sec}s");
}

sub handle_ping {
	# Don't send more than 1 ping at a time, so the lag counter works
	# properly when we're disconnected..
	return if defined $ping_last and $ping_got < $ping_last;

	$msn->ping;
	$ping_in_progress = 1;
	$ping_last = $time
}

sub handle_part {
	my $num = shift;
	$num ||= $session;

	return if $num == 0;		# XXX

	$msn->part_sess($num) or return &error
}

sub handle_window { change_win(shift) }

sub handle_destroy {
	my $num = shift;
	$num ||= $session;

	return error("Can't destroy the main window") if $num == 0;

	$msn->part_sess($num);

	change_win(0) if $session == $num;

	delete $sessions{$num}
}

sub handle_refresh { curscr->refresh }

sub handle_eval {
	my $code = shift;

	eval $code;

	error($@) if $@
}

sub handle_exec {
	my $code = shift;

	$code .= ' 2>&1';

	if($code =~ s/^\s*-o\s+//) {		# Output to chat
		return error("Can't output to main screen") if $session == 0;
		handle_msg($session, `$code`)
	} else {
		out(undef, `$code`)
	}
}

sub handle_file_send {
	my ($user, $file) = @_;
	
	$msn->file_send($user, $file) or return &error;
	out(undef, "Offered '$file' to user '$user'")
}

sub handle_file_get {
	my $num = shift;

	$msn->file_get($num, 1) or return &error;
	out(undef, "Starting file transfer $num")
}

sub handle_file_noget {
	my $num = shift;

	$msn->file_get($num, 0) or return &error;
	out(undef, "Disallowed file transfer $num")
}

sub handle_invite {
	my ($user, $sess) = @_;

	$sess ||= $session;			# XXX
	
	$msn->invite_users($sess, $user)
}

sub handle_list {
	my %users = $msn->get_users;
	my %out;

	for(keys %users) {
		my $s = $msn->get_human_state($_);
		push @{$out{$s}}, "$_ (" . get_fuser($users{$_}) . ')';
	}

	@_ = grep $_ ne 'offline', keys %out unless @_;

	my @out;
	for(@_) {
		my @t = map { "  $_" } @{$out{$_}};

		my $s = scalar(@t) . " users are currently $_:";
		$s = "1 user is currently $_:" if @t == 1;
		$s = "No users are currently $_" if !@t;

		push @out, $s, @t
	}

	out_block(undef, '', @out, '')
}

sub handle_save {
	my $file = shift;
	open T, '>', $file or return error("open of '$file' failed: $!");
	print T join "\n", @{$sessions{$session}{Screen}};
	close T;
}

sub handle_show_fuser { out(undef, get_fuser($msn->get_user(shift), 1)) }

sub handle_msg {
	my ($to, $text) = @_;

	$last_msg = $to;
	if($msn->sess_exists($to)) {
		out($to, &timestamp . '<' . $msn->plain_fuser($msn->{FUser})
								. "> $text");
		$msn->send_msg_swb($to, $text)
	}

	# XXX: Else maybe start new swb and msg them.
}

sub handle_crash {
	$msn->crash_swb(shift || $session) or out(undef, undef, 'Invalid session')
}

sub handle_help {
	my $cmd = shift;

	if(defined $cmd and $cmd =~ /(\S+)/) {	# Help on a specific command
		$cmd = $1;
		my @matches = complete($cmd, [keys %funcs]);
		if(@matches == 0) {
			return out(undef, "Invalid command '$cmd'")
		} elsif(@matches == 1) {		# Help for just this one
			my $help = $funcs{$matches[0]}[1];
			my $str = $help->[2];
			my @opts;

			# Walk through the arguments list and turn it into a
			# string...
			my ($arg, $num);
			for(@{$help->[1]}) {
				if($arg) {
					my $c = $text_translate{$_};
					$num++;

					my ($start, $end);
					my @s = split ' ', $arg;
					if(grep {$_ eq 'OPT'} @s) {
						$start = '[';
						$end = ']'
					} else {
						$start = '<';
						$end = '>'
					}

					$str =~ s/\[$num]/${start}${c}${end}/g;

					$end .= '...' if grep {$_ eq 'MUL'} @s;

					push @opts, $start . $c . $end;

					undef $arg
				} else {
					$arg = $_
				}
			}

			out(undef, "/$matches[0]", @opts);
			out(undef, $str);
			return
		}
	}

	list_commands($cmd)
}

################################################################################
# Callbacks
################################################################################

sub do_error { out(shift->get_sess_id, 'Error:', @_); event('error') }
sub do_disconnected { error('Disconnected!'); &quit }
sub do_connect { out(undef, 'Connected!'); $ping_last = 0; &handle_ping }
sub do_unknown_user { out(shift->get_sess_id, 'Unknown user', shift) }

sub do_new_mail {
	my ($msn, $msg) = @_;
	event('new_mail');
	out($msn->get_sess_id, "New mail from $msg->{From}",
				"<$msg->{FromAddr}> regarding $msg->{Subject}")
}

sub do_folder_update {
	my ($msn, $msg) = @_;
	out($msn->get_sess_id, "You have $msg->{Inbox} new messages in your inbox"
		. ($msg->{Others} ? ", and $msg->{Others} in other folders" : ''));
}

sub do_message {
	my ($msn, $msg) = @_;
	$sessions{$msn->get_sess_id}{Users}{$msg->{User}} = 0;
	event('message');
	out($msn->get_sess_id, &timestamp . '<' . get_fuser($msg) .
							"> $msg->{Body}");

	if(defined $away_msg and !$msn->{Auto_Sent}) {
		$msn->{Auto_Sent} = 1;
		handle_msg($msn->get_sess_id, "Auto message: $away_msg")
	}
}

sub do_ping {
	my ($msn, $td) = @_;

	$ping_got = time;			# Not $time, may be inaccurate
	my $l = $last_ping;
	$last_ping = $time_hires ? sprintf('%0.3f', $td) : int $td;
	$ping_in_progress = 0;
	&update_status if $l ne $last_ping
}

sub do_open_chat {
	my ($msn, $num) = @_;
	event('open_chat');
	new_win($num);
	out($num, &timestamp . 'Chat session started')
}

sub do_close_chat {
	my ($msn, $num) = @_;
	event('close_chat');
	out($num, &timestamp . 'Chat session ended');
	destroy_win($num)
}

sub do_join {
	my ($msn, $u) = @_;
	my $s = $msn->get_sess_id;
	$sessions{$s}{Users}{$u->{User}} = 0;
	event('join');
	out($s, '***', get_fuser($u), "($u->{User}) has joined")
}

sub do_part {
	my ($msn, $u) = @_;
	my $s = $msn->get_sess_id;
	delete $sessions{$s}{Users}{$u->{User}};
	event('part');
	out($s, &timestamp . '***', get_fuser($u),"($u->{User}) has left")
}

sub do_user_types {
	my ($msn, $u) = @_;
	my $s = $msn->get_sess_id;

	$sessions{$s}{Users}{$u->{User}} = time;
	make_active($s)
}

sub list_generic {
	my ($str, $msn, $u) = @_;
	out(0, get_fuser($u), "($u->{User}) $str")
}

sub list_generic_add	{ list_generic('has been ' . shift, @_) }

sub do_auth_success	{ list_generic_add('authorized', @_) }
sub do_deauth_success	{ list_generic_add('deauthorized', @_) }
sub do_add_success	{ list_generic_add('added', @_) }
sub do_del_success	{ list_generic_add('removed', @_) }
sub do_block_success	{ list_generic_add('blocked', @_) }
sub do_unblock_success	{ list_generic_add('unblocked', @_) }
sub do_del		{ list_generic('has removed you', @_) }

sub do_add {
	my ($msn, $u) = @_;
	event('add');
	list_generic('requests to be added', @_);
	out(0, "Type /auth $u->{User} to allow them to add you")
}

sub do_state_success { &update_status }
sub do_fuser_success { &update_status }

sub do_change_state {
	my ($msn, $user) = @_;

	my $hst = get_fuser($user);
	event('change_state');
        out(0, &timestamp . "$hst ($user->{User}) has changed status to '" .
				$msn->get_human_state($user->{User}) . "'")
}

sub do_out {
	my ($msn, $reason) = @_;
	out(undef, "Quitting: $reason");
	&quit
}

sub do_file_invite {
	my ($msn, $msg) = @_;

	event('file_invite');
	out($msn->get_sess_id, "*** $msg->{User} requests you to get file",
					"$msg->{File}, size $msg->{Size}");
	out($msn->get_sess_id, "Type /get $msg->{Num} to get it")
}

sub do_file_start {
	my ($msn, $msg) = @_;
	event('file_start');
	out(0, "Transfer $msg->{Num} started, IP $msg->{Host}: $msg->{Port}")
}

sub do_file_get_end {
	my ($msn, $msg) = @_;
	event('file_end');
	out(0, "Transfer $msg->{Num} (get) finished")
}

sub do_file_send_end {
	my ($msn, $msg) = @_;
	event('file_end');
	out(0, "Transfer $msg->{Num} (send) finished")
}

################################################################################
# Windowing functions
################################################################################

sub init_screen {
	initscr;			# Curses initalization
	noecho;				# Don't echo the chars we recieve

	$Text::Wrap::columns = $COLS;

	# Init colors
	start_color;
	init_pair 1, COLOR_WHITE, COLOR_MAGENTA;
	init_pair 2, COLOR_WHITE, COLOR_RED;
	init_pair 3, COLOR_WHITE, COLOR_BLUE;    

	$MAGENTA = COLOR_PAIR(1);  
	$RED     = COLOR_PAIR(2);
	$BLUE    = COLOR_PAIR(3);

	$us_line = newwin(1, 0, 0, 0);
	$us_line->leaveok(1);

	# New window for reading input from
	$in_line = newwin(1, 0, $LINES - 1, 0);
	$in_line->nodelay(1);		# Non-blocking reading
	$in_line->keypad(1);		# Enable nice character mappings

	# New window for reading input from
	$stat_line = newwin(1, 0, $LINES - 2, 0);
	$stat_line->attron($BLUE);
	$stat_line->leaveok(1);
}

sub new_win {
	my $num = shift;

	if(exists $sessions{$num}) {
		# Just add a blank line on if we have had a previous chat on
		# this window
		out($num, '')
	} else {
		$sessions{$num} = {
					Screen => [],
					Users => {},
					Scroll => 0,
					TPos => 0,
					Pos => 0,
					Lines => $LINES - 2,
				};
		my $w = newwin($sessions{$num}{Lines}, 0,
						$sessions{$num}{Pos}, 0);
		$w->scrollok(1);
		$w->leaveok(1);

		$sessions{$num}{Window} = $w;
		&update_status
	}

	$sessions{$num}{Type} = 0
}

sub change_win {
	my $num = shift;

	return unless exists $sessions{$num};
	return if $num == $session;

	$prompt = "[$num]" if !$ask;

	$session = $num;
	delete $active{$num};
	delete $vactive{$num};

	@users = ();

	# Window now contains old stuff, make sure its redrawn
	$sessions{$num}{Window}->redrawwin;

	&update_win;
	&update_status
}

sub destroy_win {
	my $win = shift;
	$sessions{$win}{Users} = {};
	&update_win if $win == $session
}

# XXX: seg faults when resized...
#sub resize_term {
#	$needs_resize = 0;
#	my ($l, $c) = ($LINES, $COLS);
#
#	# Euckie!
#	for(`$RESIZE_PROG`) {
#		/LINE.*=/i and ($LINES) = /(\d+)/;
#		/COL.*=/i and ($COLS) = /(\d+)/;
#	}
#
#	endwin;
#	initscr;
#
#	$win->resize($LINES - 2, $COLS);
#
#	$stat_line->mvwin($LINES - 2, 0);
#	$stat_line->resize(1, $COLS);
#
#	$in_line->mvwin($LINES - 1, 0);
#	$in_line->resize(1, $COLS);
#
#	&handle_refresh;
#
#	out(undef, "$l -> $LINES,  $c -> $COLS")
#}

sub out {
	out_block(shift, split "\n", wrap('', ' ' x 8, @_));
}

# immediatly?, window number, message. win == undef and puts to current window
sub out_block {
	my $num = shift;
	$num = $session if !defined $num;

	my $scr = $sessions{$num};
	my $win = $scr->{Window};

	unshift @{$scr->{Screen}}, @_?reverse @_:'';
	if(@{$scr->{Screen}} > $MAX_BUFFER_LEN) {
		splice @{$scr->{Screen}}, $MAX_BUFFER_LEN
	}

	make_vactive($num);
	if($scr->{Scroll}) {
		$scr->{Scroll} += @_
	} else {
		$win->addstr(($scr->{TPos}?"\n":'') . join "\n", @_);
		$scr->{TPos} += @_ || 1;
		$scr->{TPos} = $scr->{Lines} if $scr->{TPos} > $scr->{Lines};

		&update_win if $num == $session
	}
}

sub update_win {
	my $scr = $sessions{$session};
	my $win = $scr->{Window};
	my $n = keys %{$scr->{Users}};

	if($scr->{Pos} != $n or $us_len != $n) {
		my $d = $scr->{Pos} - $n;

		if($scr->{TPos} != $scr->{Lines} and $d != 0) {
			$win->scrl($d);
			$scr->{TPos} -= $d;
			$win->move($scr->{TPos} - 1, $COLS - 1)
		}

		$win->redrawwin if $d > 0;

		if($n > 0) {
			$us_len = $n;
			$us_line->resize($n, $COLS)
		}

		$scr->{Pos} = $n;
	}

	$last_user = $time;
	$last_active = 0;

	# Check to see if the user state/list has changed
	for(sort keys %{$scr->{Users}}) {
		$n--;
		my $col;

		my $c = $scr->{Users}{$_};
		if($c == 0) {
			$col = $BLUE
		} elsif($time > $c + 10) {
			$col = $MAGENTA
		} else {		# Timeout of 10 seconds on type
			$col = $RED;
			$last_active = $c if !$last_active or $c < $last_active
		}

		my $t = $users[$n];
		next if $t and $t->[0] eq $_ and $t->[1] == $col;
		$users[$n] = [$_, $col];

		my $str = "$_ - " . get_fuser($msn->get_user($_), 1);

		$str = substr $str, 0, $COLS;

		$us_line->move($n, 0);
		$us_line->attron($col);
		$us_line->addstr($str . ' ' x ($COLS - length $str));
		$us_line->attroff($col);
	}

	$scr->{Up} = 1
}

sub update_status {
	$stat_line->move(0,0);
	$stat_line->addstr(' ' x $COLS);
	$stat_line->move(0,0);

	if(defined $msn) {
		my $status = '';
		$status .= "[$last_ping] " if $conf{ping};
		$status .= $msn->plain_fuser($msn->{FUser}) .
					" ($msn->{User}) - " . $msn->get_state;
		$status .= " ($away_msg)" if defined $away_msg;
		$status .= ' -- more --' if $sessions{$session}{Scroll};
		$stat_line->addstr($status);
	} else {					# Initializing...
		$stat_line->addstr('Initalizing')
	}

	if(keys %active or keys %vactive) {
		$stat_line->move(0, $COLS - (keys(%active) + keys(%vactive)) * 2 - 1);
		$stat_line->addch('[');

		my $s = join ',' => sort { $a <=> $b } keys %active;
		my @v = sort { $a <=> $b } keys %vactive;

		while(defined($_ = pop @v)) {
			$stat_line->attron(A_BOLD);
			$stat_line->addstr($_);
			$stat_line->attroff(A_BOLD);
		
			$stat_line->addch(',') if @v > 0 or $s
		}

		$stat_line->addstr("$s]")
	}

	$stat_up = 1;
}

sub update_prompt {
	$in_line->erase;

	my $p = "$prompt ";
	my $l = length $p;

	$lines[$line_ptr] = '' if not defined $lines[$line_ptr];
	$p .= $lines[$line_ptr] unless $hide_input;

	if($char_ptr == -1 and length $p < $COLS) {
		$in_line->addstr($p)
	} else {
		my $curs = $char_ptr + $l;
		$curs = length $p if $char_ptr == -1;

		# Make this cleverer..
		if($curs >= $COLS) {
			$p = substr $p, 1 - $COLS + $curs, $curs;
			$curs = $COLS
		}

		$in_line->addstr($p);
		$in_line->move(0, $curs)
	}

	$in_up = 1
}

sub screen_update {
	my $scr = $sessions{$session};

	# Update the screen
	$scr->{Window}->refresh if $scr->{Up};
	if(($scr->{Up} or $user_up) and $scr->{Pos} > 0) {
		$us_line->redrawwin;
		$us_line->refresh
	}
	$stat_line->refresh if $stat_up;
	$in_line->refresh if $in_up or $stat_up or $user_up or $scr->{Up};
	$scr->{Up} = 0; $user_up = 0; $stat_up = 0;
}

################################################################################
# Input functions
################################################################################

# Returns true if the specified string is in an array
sub is_in {
	my $elem = shift;
	return (grep { $_ eq $elem } @_) ? 1 : 0
}

# first param is wether to hide the input, the rest are the message
sub ask {
	my $old_hide_input = $hide_input;
	my $old_prompt = $prompt;
	my @old_lines = @lines;
	my $old_line_ptr = $line_ptr;

	$hide_input = shift;
	$ask = 1;
	$prompt = "@_";
	@lines = ('');
	$line_ptr = 0;

	event('ask');
	&update_prompt;

	my $line;
	do {
		&screen_update;
		select $stdin, undef, undef, undef;
	} while(!defined($line = &in));

	$hide_input = $old_hide_input;
	$prompt = $old_prompt;
	$ask = 0;
	@lines = @old_lines;
	$line_ptr = $old_line_ptr;
	&update_prompt;

	$line
}

# Return a full line, or undef if not avaliable (non-blocking)
sub in {
	while((my $t = $in_line->getch) ne ERR) {
		# Auto-idle stuff
		if($idle) {
			$idle = 0;
			$msn->change_my_state('online');
			out(undef, 'Unidleing after being idle for',
					int($time - $last_type), 'seconds')
		}

		$last_type = $time;

		# Quickly handle common ones first:
		if($escape) {				# Alt-command
			$escape = 0;

			if($t =~ /^\d$/) {		# alt-num, change window
				change_win(($t || 10) - 1)
			} elsif($t eq KEY_BACKSPACE or $t eq "\b") {
				my ($end, $beg) = ('');
				if($char_ptr == -1) {
					$beg = $lines[$line_ptr]
				} else {
					$beg = substr $lines[$line_ptr], 0,
								$char_ptr;
					$end = substr $lines[$line_ptr],
							$char_ptr,
							length $lines[$line_ptr]
				}

				$beg =~ s/([\w\d]*[^\w\d]*)$//;

				$lines[$line_ptr] = "$beg$end";
				$char_ptr = length $beg if $char_ptr != -1
			}
		} elsif($t =~ /^[\040-\377]$/s) {
			# Send a user-is-typing message to the swb if we havn't
			# in the last 4 seconds.
			if($session and $time >= $sessions{$session}{Type} + 4){
				$sessions{$session}{Type} = $time;
				$msn->send_typing_swb($session)
			}

			if($char_ptr == -1) {
				$lines[$line_ptr] .= $t
			} else {
				my $len = $replace_on ? 1 : 0;
				substr $lines[$line_ptr], $char_ptr++, $len, $t
			}
		} elsif($t eq "\n") {
			next if $lines[$line_ptr] eq '';
			$lines[0] = $t = $lines[$line_ptr];

			$char_ptr = -1;
			$line_ptr = 0;

			unshift @lines, '';

			pop @lines if @lines > $MAX_LINE_LEN;

			# Reset the time we last typed.
			$sessions{$session}{Type} = 0;

			&update_prompt;

			return $t
		} elsif($t eq "\t") {{			# Tab-completion
			last if $ask;

			if($lines[$line_ptr] =~ /^\s*$/) {
				# Just a <tab>, recall last msg command
				$lines[$line_ptr] = "/msg $last_msg ";

				# char_ptr == -1 already
				last
			}

			my ($end, $beg) = ('');
			if($char_ptr == -1) {
				$beg = $lines[$line_ptr]
			} else {
				$beg = substr $lines[$line_ptr], 0, $char_ptr;
				$end = substr $lines[$line_ptr], $char_ptr,
						length $lines[$line_ptr];
			}

			# Get the current element we are on
			my $pos = 0;
			$pos++ while $beg =~ /(\s+)/cg;

			# If we are the first thing, we must be a command
			my $type = 'COMMAND';

			# We now find the type of this command
			if($pos > 0) {
				my $command;
				$beg =~ /^\/(\S*)/ or last;
				do_cmd(\$command, $1) or last;

				my $args = $funcs{$command}[1][1];

				# Hm, invalid args
				if(!@$args) {
					event('command');
					last
				} elsif($pos > @$args / 2) {
					# if it is part the end of the list
					# Last option should be MUL
					if($args->[-2] !~ /\bMUL\b/) {
						event('command');
						last
					}

					$type = $args->[-1]
				} else {
					$type = $args->[$pos * 2 - 1]
				}
			}

			# $inc is the word we want to complete
			$beg =~ s/(\S*)$//;
			my $inc = $1;

			my ($comp, $r);

			# Switch through the types now
			if($type eq 'COMMAND') {
				my $s;
				$inc =~ s/^\/// and $s++;

				$r = do_cmd(\$comp, $inc);
				last if !defined $comp;

				$comp = ($s ? '/' : '') . $comp
			} elsif($type eq 'USER') {
				my %t = $msn->get_users;
				$r = do_comp(\$comp, $inc, [keys %t]);
			} elsif($type eq 'STATE') {
				$r = do_comp(\$comp, $inc, [keys %MSN::STATES])
			} elsif($type eq 'FILE') {
				my @files;
				for(glob "$inc*") { push @files, $_ if -f }
				$r = do_comp(\$comp, $inc, \@files)
			} elsif($type eq 'WINDOW') {
				$r = do_comp(\$comp, $inc, [keys %sessions])
			} elsif($type eq 'SESSION') {
				my $sess = [];

				for(keys %sessions) {
					next unless $msn->sess_exists($_);
					push @$sess, $_ if $_
				}

				$r = do_comp(\$comp, $inc, $sess);
			} elsif($type eq 'TEXT') {
				# Do nothing
			} else {
				error("Client does not support type $type");
				last
			}

			last unless defined $r and defined $comp;
			$r == 1 and $comp .= ' ';
			$beg .= $comp;

			$lines[$line_ptr] = "$beg$end";
			$char_ptr = length $beg if $char_ptr != -1
		}} elsif($t eq "\e") {			# Escape
			$escape = 1

		} elsif($t eq KEY_BACKSPACE or $t eq "\b") {
			if($char_ptr == -1) {
				chop $lines[$line_ptr]
			} else {
				substr $lines[$line_ptr], --$char_ptr, 1, ''
								if $char_ptr > 0
			}
		} elsif($t eq KEY_DC) {
			substr $lines[$line_ptr], $char_ptr, 1, ''
							if $char_ptr != -1
		} elsif($t eq KEY_UP) {
			$line_ptr = 0 if ++$line_ptr >= @lines;
			$char_ptr = -1
		} elsif($t eq KEY_DOWN) {
			$line_ptr = 0 if --$line_ptr <= -@lines;
			$char_ptr = -1
		} elsif($t eq KEY_LEFT) {
			if($char_ptr == -1) {
				# Clever, if length == 0, char_ptr = -1
				$char_ptr = length($lines[$line_ptr]) - 1
			} else {
				$char_ptr-- if $char_ptr > 0
			}
		} elsif($t eq KEY_RIGHT) {
			my $v = length $lines[$line_ptr];
			if($char_ptr != -1 and $char_ptr < $v) {
				$char_ptr = -1 if ++$char_ptr == $v
			}
		} elsif($t eq KEY_HOME) {
			$char_ptr = 0
		} elsif($t eq KEY_END) {
			$char_ptr = -1
		} elsif($t eq KEY_PPAGE or $t eq KEY_NPAGE) {
			my $scr = $sessions{$session};
			my $win = $scr->{Window};
			my $l = $scr->{Lines};
			my $sc = $scr->{Screen};
			my $s = \$scr->{Scroll};
			my $o = $$s;
			my $pos = 0;
			my @arr;

			# Not filled a page yet	or scrollback buffer too small
			next if $#$sc <= $l;
			
			if($t eq KEY_PPAGE and $o < $MAX_BUFFER_LEN) {
				$$s += int($l / 2);
				$$s = @$sc - $l if $$s > @$sc - $l;
				@arr = @{$sc}[$o + $l .. $$s + $l - 1]
			} elsif($t eq KEY_NPAGE and $o > 0) {
				if($o > $MAX_BUFFER_LEN - $l) {
					$$s = $MAX_BUFFER_LEN - $l
				} else {
					$$s -= int($l / 2)
				}

				$$s = 0 if $$s < 0;
				@arr = @{$sc}[$$s .. $o - 1];
				$pos = $l + $$s - $o
			} else {
				next
			}

			if($o - $$s >= $l) {
				$win->move(0,0);
				@arr = @{$sc}[$$s .. $$s + $l - 1];
			} else {
				next if $o == $$s;
				$win->scrl($o - $$s);
				$win->move($pos, 0);
			}

			$win->addstr(join "\n", reverse @arr);
			&update_win;
			&update_status
		} elsif($t eq KEY_IC) {
			$replace_on = !$replace_on;
			next
		} elsif($t eq "\014") {			# ^L
			&handle_refresh;
			next
		}

		&update_prompt
	}

	undef
}

################################################################################
# Useful functions
################################################################################

# this does some completion magic or panics about there being multiple or no
# completions
sub do_intern_comp {
	my ($ret, $func, $thing, $things) = @_;

	my @things = complete($thing, $things);

	if(@things == 1) {
		$$ret = $things[0];
		return 1
	} elsif(@things) {
		# Now we look for the longest string common between the matches
		# nb, only compare arrays on length
		my $len = length $thing;
		1 while complete(substr($things[0], 0, ++$len), \@things)
								== @things;

		$$ret = substr $things[0], 0, $len if --$len != length $thing;

		event('multmatch');

		&$func(@things);

		return 0
	} else {
		event('nomatch');
		out(undef, 'No match found for', $thing) if $thing ne ''
	}

	undef
}

sub do_comp {
	do_intern_comp(shift,
		sub { out_block(undef, 'Matches are:', map {"  $_"} @_) },@_)
}

sub do_cmd {
	my ($r, $c) = @_;
	do_intern_comp($r, sub { list_commands($c) }, $c, [keys %funcs])
}

# Returns the timestamp in the format the luser wanted it in :-)
sub timestamp {
	return strftime($conf{time}, localtime) . ' ' if $conf{time};
	''
}

sub read_config {
	# Try to read the config
	if(open G, $CONFIG) {
		my $config_file;
		while(<G>) {
			chomp;
			s/\s*#.*//g;
			next unless $_;
			$config_file .= "$_,"
		}
		close G;
		eval "\%conf = (\%conf, $config_file)";

		# Process beeping section
		$conf{beep_} = { map { $_ => 1 } split ' ', $conf{beep} };
	} else {
		print "Warning: Can't open $CONFIG: $!"
	}
}

sub make_active {
	my $num = shift;
	return &update_win if $num == $session;
	return if exists $active{$num} or exists $vactive{$num};
	$active{$num}++;
	&update_status
}

sub make_vactive {
	my $num = shift;
	return if $num == $session or exists $vactive{$num};
	delete $active{$num};
	$vactive{$num}++;
	&update_status
}

sub get_fuser {
	my $h = shift;
	my $nick = $msn->plain_fuser($msn->get_user($h->{User})->{FUser});
	$nick = substr $nick, 0, $MAX_FUSER_LEN if $MAX_FUSER_LEN and !shift;
	$nick
}

# Returns a command from a list
sub complete {
	my ($word, $words) = @_;

	return @$words unless defined $word;

	# First check complete matches
	for(@$words) {
		return (@_ = $_) if lc $_ eq lc $word
	}

	# Then check for partial matches and return
	@_ = grep /^\Q$word\E/i, @$words;
}

sub list_commands {
	my $start = shift;

	$start ||= '';

	my %uniq;

	$uniq{$funcs{$_}[0]}++ for complete($start, [keys %funcs]);

	return out(undef, "No matches found for $start") unless keys %uniq;

	my @l;

	for my $t (map {$funcs{$_}[1]} sort keys %uniq) {
		my @commands = @{$t->[0]};
		if(@commands == 1) {			# 1 command
			push @l, $commands[0]
		} else {				# Multiple commands
			push @l, "{@commands}"
		}
	}

	out_block(undef, 'Matches are:', map {"  $_"} @l);
}

sub event {
	my $event = shift;
	beep if exists $conf{beep_}{$event}
}

sub error {
	my $text = "@_";

	event('error');

	$text ||= $msn->{Error};
	out(undef, "Error: $text")
}
